/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/****  Project : eXtended C Library (XCLIB)
    *  Purpose : CIRCULAR DOUBLY LINKED LIST WITH DUMMY NODE
    *  Language: ANSI C
    *  File    : demo_CLListD-1.c
    *  Author  : Martin Vicente <martin@vicente.org>
    *
    ****/


#include <stddef.h>
#include <stdio.h>
#include <assert.h>

#include <xc/CLListD.h>

#undef	__CLLD_LINK
#define	__CLLD_LINK(LIST)	(&(LIST)->CLLD_link)

typedef struct node *Node;
typedef struct list *List;

struct node
{
    struct node *next;
    struct node *prev;

    int data;
};

struct list
{
    struct node link; /* sentinel */
};

#define L1	"List: ("
#define L2	" )\n"

static void print(const struct list *list);
static void put(const char *s);

#if 1
#define TRACE__(item)	_MBEGIN puts("\
----------------------------------------------------------------------\r\
-------------------- " #item " "); item; _MEND
#else
#define TRACE__(item)	item
#endif

#define COMM(en, fr)	__xcint(comm)(en, fr, NULL)
#define WAIT()		__xcint(wait)()

extern void __xcint(comm)(const char *en, ...);	/* DON'T USE IT!!! */
extern void __xcint(wait)(void);		/* DON'T USE IT!!! */

int main()
{
    struct list  list;
    struct node  node1;
    struct node  node2;
    struct node  node3;
    struct node  node4;
    struct node  node5;
    size_t       len;
    struct node *n;

    COMM("CIRCULAR DOUBLY LINKED LIST WITH DUMMY NODE",
	 "LISTE CIRCULAIRE DOUBLEMENT CHAINEE AVEC NOEUD FACTICE");

    TRACE__(CLLD_INIT(&list));
    print(&list);

    node1.data = 1;
    TRACE__(CLLD_APPEND(&list, &node1));
    print(&list);

    node2.data = 2;
    TRACE__(CLLD_APPEND(&list, &node2));
    print(&list);

    node3.data = 3;
    TRACE__(CLLD_APPEND(&list, &node3));
    print(&list);

    node4.data = 4;
    TRACE__(CLLD_INSERT(&list, &node4));
    print(&list);

    node5.data = 5;
    TRACE__(CLLD_INSERT(&list, &node5));
    print(&list);

    TRACE__(CLLD_PLENGTH(&list, len));
    printf("length = %lu\n", (unsigned long int)len);

    TRACE__(CLLD_REMOVE(&node5));
    print(&list);

    TRACE__(CLLD_INSERT_AFTER(&node5, &node4));
    print(&list);

    TRACE__(CLLD_REMOVE(&node3));
    print(&list);

    TRACE__(CLLD_INSERT_BEFORE(&node3, &node2));
    print(&list);

#if defined __GNUC__ && !defined __STRICT_ANSI__
    TRACE__(n = CLLD_SELECT(&list, 0));
#else
    TRACE__(CLLD_SELECT(&list, 0, n));
#endif
    TRACE__(CLLD_REMOVE(n));
    TRACE__(CLLD_APPEND(&list, n));
    print(&list);

    WAIT();

    return 0;
}

static void print(const struct list *const list)
{
    const struct node *n;

    assert(list != NULL);

    put(L1);

#if 0
    CLLD_FOREACH_NODE(list, printf(" %d", __node->data); );
#else
    CLLD_FOREACH(list, n) printf(" %d", n->data);
#endif

    put(L2);

    if ((n = CLLD_FIRST(list)) != NULL) printf("First: %d\n", n->data);
    if ((n = CLLD_LAST(list))  != NULL) printf("Last:  %d\n", n->data);
}

static void put(const char *s)
{
    fputs(s, stdout);
}


/* End of file */
