/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/****  Project : eXtended C Library (XCLIB)
    *  Purpose : DOUBLY LINKED LIST WITH DUMMY NODES
    *  Language: ANSI C
    *  File    : demo_LListD-3.c
    *  Author  : Martin Vicente <martin@vicente.org>
    *
    ****/


#include <stdio.h>

#include <xc/LListD.h>
#include <xc/xalloc.h>

typedef struct node *Node;
typedef struct list *List;

struct node
{
    Node next;
    Node prev;
    int  data;
};

struct list
{
    Node head;
    Node tail;
};

int main()
{
    List l;
    Node n;
    int  i;

    l = xnew(List);
    l->head = xnew(Node);
    l->tail = xnew(Node);
    LLD_INIT(l);

    for (i = 1; i <= 10; i++)
    {
	(n = xnew(Node))->data = i;

	LLD_APPEND(l, n);
    }

    LLD_FWD_ACTION(l, printf(" %02d", __node->data); );
    putchar('\n');
    LLD_BWD_ACTION(l, printf(" %02d", __node->data); );
    putchar('\n');

    for (n = LLD_FIRST(l); n; n = LLD_NEXT(l, n)) printf(" %02d", n->data);
    putchar('\n');

    for (n = LLD_LAST(l); n; n = LLD_PREV(l, n)) printf(" %02d", n->data);
    putchar('\n');

    LLD_CLEAN(l, xfree(__node); );

    xfree(l->head);
    xfree(l->tail);
    xfree(l);

    return 0;
}


/* End of file */
