/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*  XCLIB 0.3	CLIST:		CIRCULAR SINGLY LINKED LIST		     */


#ifndef __XC_CList_H
#define __XC_CList_H


/* Standard C Library: */
#include <stddef.h>
#include <assert.h>

/* eXtended C Library: */
#include <xc/version.h>		/* VERSION OF EXTENDED C LIBRARY	     */


/*****************************************************************************
 * CUSTOMIZATION PART
 *****************************************************************************/


#ifndef CL_NodeRec
#define CL_NodeRec  CL_NodeRec
typedef struct node CL_NodeRec;
#endif

#ifndef CL_ListRec
#define CL_ListRec  CL_ListRec
typedef struct list CL_ListRec;
#endif

#ifndef CL_Node
#define CL_Node     CL_Node
typedef CL_NodeRec *CL_Node;
#endif

#ifndef CL_List
#define CL_List     CL_List
typedef CL_ListRec *CL_List;
#endif


#ifndef CL_next
#define CL_next	next
#endif

#ifndef CL_link
#define CL_link	link
#endif


/*****************************************************************************
 * INTERNALS
 *****************************************************************************/


#if defined __GNUC__ && !defined __STRICT_ANSI__
#define CL_TYPEREC(NODEREC)	typeof (NODEREC)
#define CL_TYPE(NODE)		typeof (NODE)
#else
#define CL_TYPEREC(NODEREC)	CL_NodeRec
#define CL_TYPE(NODE)		CL_Node
#endif


#define	__CL_NEXT(NODE)	((NODE)->CL_next)
#define	__CL_LINK(LIST)	((LIST)->CL_link)


#define __CL_FIRST(L)	(L)


#define __CL_SANITY_CHECK(L) assert					      \
									      \
 ( (L) == NULL || (							      \
									      \
 __CL_NEXT(L) != NULL ) )


#define _CL_INIT(L)							      \
_MBEGIN									      \
    (L) = NULL;								      \
_MEND

#define _CL_FOREACH_COND(L, NODE, COND)					      \
									      \
 for (									      \
									      \
 assert((L) != NULL), __CL_SANITY_CHECK(L),				      \
									      \
 (NODE) = __CL_NEXT(__CL_FIRST(L)); (NODE) != (L) && (COND); (NODE) = __CL_NEXT(NODE))

#define _CL_FOREACH(L, NODE)	_CL_FOREACH_COND(L, NODE, 1)

#define _CL_FOREACH_NODE(L, INST)					      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) __node;					      \
									      \
    _CL_FOREACH(L, __node) { INST }					      \
_MEND

#define _CL_NEXT(N)	__CL_NEXT(N)

#define __CL_PREV(L, NODE, PREV)					      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) ____n;					      \
									      \
    (PREV) = (L);							      \
									      \
    _CL_FOREACH_COND(L, ____n, ____n != (NODE)) (PREV) = ____n;		      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _CL_PREV(L, NODE)						      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) ___prev;					      \
									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    __CL_PREV(L, NODE, ___prev);					      \
									      \
    ___prev;								      \
_MEND
#else
#define _CL_PREV(L, NODE, PREV)						      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) ___prev;					      \
									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    __CL_PREV(L, NODE, ___prev);					      \
									      \
    (PREV) = ___prev;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_CL_PPREV(L, NODE, PREV)	((PREV) = _CL_PREV(L, NODE))
#else
#define	_CL_PPREV(L, NODE, PREV)	_CL_PREV(L, NODE, PREV)
#endif

#define _CL_FIRST(L)	(__CL_SANITY_CHECK(L), __CL_FIRST(L))

#define	__CL_LAST(L, LAST)						      \
_MBEGIN									      \
    if ((L) != NULL) __CL_PREV(L, L, LAST); else (LAST) = NULL;		      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_CL_LAST(L)							      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) ___last_node;				      \
									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    __CL_LAST(L, ___last_node);						      \
									      \
    ___last_node;							      \
_MEND
#else
#define	_CL_LAST(L, LAST)						      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) ___last_node;				      \
									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    __CL_LAST(L, ___last_node);						      \
									      \
    (LAST) = ___last_node;						      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_CL_PLAST(L, LAST)	((LAST) = _CL_LAST(L))
#else
#define	_CL_PLAST(L, LAST)	_CL_LAST(L, LAST)
#endif

#define _CL_INSERT_BEFORE(L, NODE, MARK)				      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) ___prev_mark;				      \
									      \
    assert((NODE) != NULL);						      \
    assert((MARK) != NULL);						      \
									      \
    __CL_NEXT(NODE) = (MARK);						      \
									      \
    __CL_PREV(L, MARK, ___prev_mark);					      \
									      \
    __CL_NEXT(___prev_mark) = (NODE);					      \
_MEND

#define _CL_INSERT_AFTER(NODE, MARK)					      \
_MBEGIN									      \
    assert((NODE) != NULL);						      \
    assert((MARK) != NULL);						      \
									      \
    __CL_NEXT(NODE) = __CL_NEXT(MARK);					      \
									      \
    __CL_NEXT(MARK) = (NODE);						      \
_MEND

#define _CL_APPEND(L, NODE)						      \
_MBEGIN									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    if ((L) != NULL)							      \
    {									      \
	_CL_INSERT_BEFORE(L, NODE, L);					      \
    }									      \
    else								      \
    {									      \
	__CL_NEXT(NODE) = (NODE);					      \
									      \
	(L) = (NODE);							      \
    }									      \
_MEND

#define _CL_INSERT(L, NODE)						      \
_MBEGIN									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    if ((L) != NULL)							      \
    {									      \
	_CL_INSERT_BEFORE(L, NODE, L);					      \
    }									      \
    else								      \
    {									      \
	__CL_NEXT(NODE) = (NODE);					      \
    }									      \
									      \
    (L) = (NODE);							      \
_MEND

#define _CL_PREPEND(L, NODE) _CL_INSERT(L, NODE)

#define __CL_REMOVE(L, NODE)						      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(NODE) ___prev_node;				      \
									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    assert((NODE) != NULL);						      \
									      \
    if ((L) == (NODE)) (L) = __CL_NEXT(NODE);				      \
    if ((L) == (NODE)) (L) = NULL;					      \
									      \
    __CL_PREV(L, NODE, ___prev_node);					      \
									      \
    __CL_NEXT(___prev_node) = __CL_NEXT(NODE);				      \
									      \
    __CL_NEXT(NODE) = NULL;						      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _CL_REMOVE(L, NODE)	(__CL_REMOVE(L, NODE), (NODE))
#else
#define _CL_REMOVE(L, NODE)	__CL_REMOVE(L, NODE)
#endif

#define _CL_EXTRACT(L, NODE) _CL_REMOVE(L, NODE)

#define _CL_FOREWARD_ACTION(L, INST)					      \
_MBEGIN									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    if ((L) != NULL)							      \
    {									      \
	__XC_REGISTER CL_TYPE(L) ___first_node, __node, ___next_node;	      \
									      \
	___first_node = __CL_FIRST(L);					      \
									      \
	__node = ___first_node;						      \
									      \
	___next_node = __CL_NEXT(__node);				      \
									      \
	INST								      \
									      \
	for (__node = ___next_node; __node != ___first_node; __node = ___next_node)\
	{								      \
	    ___next_node = __CL_NEXT(__node);				      \
									      \
	    INST							      \
	}								      \
    }									      \
_MEND

#define _CL_BACKWARD_ACTION(L, INST)					      \
_MBEGIN									      \
    __CL_SANITY_CHECK(L);						      \
									      \
    if ((L) != NULL)							      \
    {									      \
	__XC_REGISTER CL_TYPE(L) ___last_node, __node, ___prev_node;	      \
									      \
	__CL_LAST(L, ___last_node);					      \
									      \
	__node = ___last_node;						      \
									      \
	__CL_PREV(L, __node, ___prev_node);				      \
									      \
	INST								      \
									      \
	for (__node = ___prev_node; __node != ___last_node; __node = ___prev_node)\
	{								      \
	    __CL_PREV(L, __node, ___prev_node);				      \
									      \
	    INST							      \
	}								      \
    }									      \
_MEND

#define _CL_ACTION(L, INST)	_CL_FOREWARD_ACTION(L, INST)
#define _CL_FWD_ACTION(L, INST)	_CL_FOREWARD_ACTION(L, INST)
#define _CL_BWD_ACTION(L, INST)	_CL_BACKWARD_ACTION(L, INST)

#define _CL_CLEAN(L, INST)						      \
_MBEGIN									      \
    _CL_ACTION(L, INST);						      \
									      \
    _CL_INIT(L);							      \
_MEND

#define _CL_PURGE(L, INST)	_CL_PURGE(L, INST)

#define __CL_LENGTH(L, LEN)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __i = 0;					      \
									      \
    if ((L) != NULL) { __i++; _CL_FOREACH_NODE(L, __i++;); }		      \
									      \
    (LEN) = __i;							      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _CL_LENGTH(L)							      \
_MBEGIN									      \
    __XC_REGISTER size_t __len;						      \
									      \
    __CL_LENGTH(L, __len);						      \
									      \
    __len;								      \
_MEND
#else
#define _CL_LENGTH(L, LEN)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __len;						      \
									      \
    __CL_LENGTH(L, __len);						      \
									      \
    (LEN) = __len;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _CL_PLENGTH(L, LEN)	((LEN) = _CL_LENGTH(L))
#else
#define _CL_PLENGTH(L, LEN)	_CL_LENGTH(L, LEN)
#endif

#define __CL_SELECT(L, INDEX, NODE)					      \
_MBEGIN									      \
    __XC_REGISTER size_t const __i = (INDEX);				      \
    __XC_REGISTER size_t       __j = 0;					      \
									      \
    if ((L) != NULL)							      \
    {									      \
	do								      \
	{								      \
	    (NODE) = __CL_FIRST(L);					      \
									      \
	    if (__j++ == __i) break;					      \
									      \
	    _CL_FOREACH(L, NODE) if (__j++ == __i) break;		      \
	}								      \
	while (0);							      \
    }									      \
    else								      \
    {									      \
	(NODE) = NULL;							      \
    }									      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _CL_SELECT(L, INDEX)						      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) __node;					      \
									      \
    __CL_SELECT(L, INDEX, __node);					      \
									      \
    __node;								      \
_MEND
#else
#define _CL_SELECT(L, INDEX, NODE)					      \
_MBEGIN									      \
    __XC_REGISTER CL_TYPE(L) __node;					      \
									      \
    __CL_SELECT(L, INDEX, __node);					      \
									      \
    (NODE) = __node;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _CL_PSELECT(L, INDEX, NODE)	((NODE) = _CL_SELECT(L, INDEX))
#else
#define _CL_PSELECT(L, INDEX, NODE)	_CL_SELECT(L, INDEX, NODE)
#endif


/*****************************************************************************
 * HIGH LEVEL MACROS
 *****************************************************************************/


#define _CL_LINK(LIST)	__CL_LINK((assert((LIST) != NULL), LIST))


#define CL_INIT(LIST)				_CL_INIT(_CL_LINK(LIST))
#define	CL_NEXT(LIST, MARK)			_CL_NEXT(MARK)
#define CL_FIRST(LIST)				_CL_FIRST(_CL_LINK(LIST))
#define CL_APPEND(LIST, NODE)			_CL_APPEND(_CL_LINK(LIST), NODE)
#define CL_INSERT(LIST, NODE)			_CL_INSERT(_CL_LINK(LIST), NODE)
#define CL_PREPEND(LIST, NODE)			_CL_PREPEND(_CL_LINK(LIST), NODE)
#define CL_INSERT_BEFORE(LIST, NODE, MARK)	_CL_INSERT_BEFORE(_CL_LINK(LIST), NODE, MARK)
#define CL_INSERT_AFTER(NODE, MARK)		_CL_INSERT_AFTER(NODE, MARK)
#define CL_REMOVE(LIST, NODE)			_CL_REMOVE(_CL_LINK(LIST), NODE)
#define CL_EXTRACT(LIST, NODE)			_CL_EXTRACT(_CL_LINK(LIST), NODE)
#define CL_FOREACH_NODE(LIST, INST)		_CL_FOREACH_NODE(_CL_LINK(LIST), INST)
#define CL_FOREACH(LIST, NODE)			_CL_FOREACH(_CL_LINK(LIST), NODE)
#define CL_FOREACH_COND(LIST, NODE, COND)	_CL_FOREACH_COND(_CL_LINK(LIST), NODE, COND)
#define CL_ACTION(LIST, INST)			_CL_ACTION(_CL_LINK(LIST), INST)
#define CL_FWD_ACTION(LIST, INST)		_CL_FWD_ACTION(_CL_LINK(LIST), INST)
#define CL_BWD_ACTION(LIST, INST)		_CL_BWD_ACTION(_CL_LINK(LIST), INST)
#define CL_FOREWARD_ACTION(LIST, INST)		_CL_FOREWARD_ACTION(_CL_LINK(LIST), INST)
#define CL_BACKWARD_ACTION(LIST, INST)		_CL_BACKWARD_ACTION(_CL_LINK(LIST), INST)
#define CL_CLEAN(LIST, INST)			_CL_CLEAN(_CL_LINK(LIST), INST)
#define CL_PURGE(LIST, INST)			_CL_PURGE(_CL_LINK(LIST), INST)

/*
 * Non portable versions
 */

#if defined __GNUC__ && !defined __STRICT_ANSI__

/* ====> Function behaviour */

#define CL_PREV(LIST, MARK)			_CL_PREV(_CL_LINK(LIST), MARK)
#define CL_LAST(LIST)				_CL_LAST(_CL_LINK(LIST))
#define CL_LENGTH(LIST)				_CL_LENGTH(_CL_LINK(LIST))
#define CL_SELECT(LIST, I)			_CL_SELECT(_CL_LINK(LIST), I)

#else

/* ====> strict ANSI conformance */

#define CL_PREV(LIST, MARK, PREV)		_CL_PREV(_CL_LINK(LIST), MARK, PREV)
#define CL_LAST(LIST, LAST)			_CL_LAST(_CL_LINK(LIST), LAST)
#define CL_LENGTH(LIST, LEN)			_CL_LENGTH(_CL_LINK(LIST), LEN)
#define CL_SELECT(LIST, I, NODE)		_CL_SELECT(_CL_LINK(LIST), I, NODE)

#endif

/*
 * Portable versions
 */

#define CL_PPREV(LIST, MARK, PREV)		_CL_PPREV(_CL_LINK(LIST), MARK, PREV)
#define CL_PLAST(LIST, LAST)			_CL_PLAST(_CL_LINK(LIST), LAST)
#define CL_PLENGTH(LIST, LEN)			_CL_PLENGTH(_CL_LINK(LIST), LEN)
#define CL_PSELECT(LIST, I, NODE)		_CL_PSELECT(_CL_LINK(LIST), I, NODE)


#endif /* XC CList HEADER */


/*------------------------------/ END OF FILE /------------------------------*/
