/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*  XCLIB 0.3	LLISTD:		DOUBLY LINKED LIST WITH DUMMY NODES	     */


#ifndef __XC_LListD_H
#define __XC_LListD_H


/* Standard C Library: */
#include <stddef.h>
#include <assert.h>

/* eXtended C Library: */
#include <xc/version.h>		/* VERSION OF EXTENDED C LIBRARY	     */


/*****************************************************************************
 * CUSTOMIZATION PART
 *****************************************************************************/


#ifndef LLD_NodeRec
#define LLD_NodeRec LLD_NodeRec
typedef struct node LLD_NodeRec;
#endif

#ifndef LLD_ListRec
#define LLD_ListRec LLD_ListRec
typedef struct list LLD_ListRec;
#endif

#ifndef LLD_Node
#define LLD_Node     LLD_Node
typedef LLD_NodeRec *LLD_Node;
#endif

#ifndef LLD_List
#define LLD_List     LLD_List
typedef LLD_ListRec *LLD_List;
#endif


#ifndef LLD_next
#define LLD_next	next
#endif

#ifndef LLD_prev
#define LLD_prev	prev
#endif

#ifndef LLD_head
#define LLD_head	head
#endif

#ifndef LLD_tail
#define LLD_tail	tail
#endif


/*****************************************************************************
 * INTERNALS
 *****************************************************************************/


#if defined __GNUC__ && !defined __STRICT_ANSI__
#define LLD_TYPEREC(NODEREC)	typeof (NODEREC)
#define LLD_TYPE(NODE)		typeof (NODE)
#else
#define LLD_TYPEREC(NODEREC)	LLD_NodeRec
#define LLD_TYPE(NODE)		LLD_Node
#endif


#define	__LLD_NEXT(NODE)	((NODE)->LLD_next)
#define	__LLD_PREV(NODE)	((NODE)->LLD_prev)
#define	__LLD_HEAD(LIST)	((LIST)->LLD_head)
#define	__LLD_TAIL(LIST)	((LIST)->LLD_tail)


#define __LLD_FIRST(H, T)	__LLD_NEXT(H)
#define __LLD_LAST(H, T)	__LLD_PREV(T)


#define __LLD_SANITY_CHECK(H, T) assert					      \
									      \
 ( (H) != NULL && (T) != NULL && (					      \
									      \
 __LLD_NEXT(H) != NULL && __LLD_PREV(H) == NULL &&			      \
 __LLD_PREV(T) != NULL && __LLD_NEXT(T) == NULL ) )


#define _LLD_INIT(H, T)							      \
_MBEGIN									      \
    assert((H) != NULL && (T) != NULL);					      \
									      \
    __LLD_NEXT(H) = (T);						      \
    __LLD_PREV(H) = NULL; /* unused */					      \
									      \
    __LLD_PREV(T) = (H);						      \
    __LLD_NEXT(T) = NULL; /* unused */					      \
_MEND

#define _LLD_FOREACH_COND(H, T, NODE, COND)				      \
									      \
 for (									      \
									      \
 __LLD_SANITY_CHECK(H, T),						      \
									      \
 (NODE) = __LLD_FIRST(H, T); (NODE) != (T) && (COND); (NODE) = __LLD_NEXT(NODE))

#define _LLD_FOREACH(H, T, NODE)	_LLD_FOREACH_COND(H, T, NODE, 1)

#define _LLD_FOREACH_NODE(H, T, INST)					      \
_MBEGIN									      \
    __XC_REGISTER LLD_TYPE(H) __node;					      \
									      \
    _LLD_FOREACH(H, T, __node) { INST }					      \
_MEND

#define _LLD_NEXT(H, T, N)	(__LLD_NEXT(N) != (T) ? __LLD_NEXT(N) : NULL)
#define _LLD_PREV(H, T, N)	(__LLD_PREV(N) != (H) ? __LLD_PREV(N) : NULL)

#define _LLD_FIRST(H, T)						      \
									      \
 (__LLD_SANITY_CHECK(H, T), __LLD_FIRST(H, T) != (T) ? __LLD_FIRST(H, T) : NULL)

#define _LLD_LAST(H, T)							      \
									      \
 (__LLD_SANITY_CHECK(H, T), __LLD_LAST(H, T)  != (H) ? __LLD_LAST (H, T) : NULL)

#define _LLD_APPEND(H, T, NODE)						      \
_MBEGIN									      \
    __LLD_SANITY_CHECK(H, T);						      \
									      \
    _LLD_INSERT_BEFORE(NODE, T);					      \
_MEND

#define _LLD_INSERT(H, T, NODE)						      \
_MBEGIN									      \
    __LLD_SANITY_CHECK(H, T);						      \
									      \
    _LLD_INSERT_AFTER(NODE, H);						      \
_MEND

#define _LLD_PREPEND(H, T, NODE) _LLD_INSERT(H, T, NODE)

#define _LLD_INSERT_BEFORE(NODE, MARK)					      \
_MBEGIN									      \
    assert((NODE) != NULL);						      \
    assert((MARK) != NULL);						      \
									      \
    __LLD_NEXT(NODE) = (MARK);						      \
    __LLD_PREV(NODE) = __LLD_PREV(MARK);				      \
									      \
    __LLD_PREV(__LLD_NEXT(NODE)) = (NODE);				      \
    __LLD_NEXT(__LLD_PREV(NODE)) = (NODE);				      \
_MEND

#define _LLD_INSERT_AFTER(NODE, MARK)					      \
_MBEGIN									      \
    assert((NODE) != NULL);						      \
    assert((MARK) != NULL);						      \
									      \
    __LLD_PREV(NODE) = (MARK);						      \
    __LLD_NEXT(NODE) = __LLD_NEXT(MARK);				      \
									      \
    __LLD_PREV(__LLD_NEXT(NODE)) = (NODE);				      \
    __LLD_NEXT(__LLD_PREV(NODE)) = (NODE);				      \
_MEND

#define __LLD_REMOVE(NODE)						      \
_MBEGIN									      \
    assert((NODE) != NULL);						      \
									      \
    __LLD_NEXT(__LLD_PREV(NODE)) = __LLD_NEXT(NODE);			      \
    __LLD_PREV(__LLD_NEXT(NODE)) = __LLD_PREV(NODE);			      \
									      \
    __LLD_NEXT(NODE) = NULL;						      \
    __LLD_PREV(NODE) = NULL;						      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _LLD_REMOVE(NODE)	(__LLD_REMOVE(NODE), (NODE))
#else
#define _LLD_REMOVE(NODE)	__LLD_REMOVE(NODE)
#endif

#define _LLD_EXTRACT(NODE)	_LLD_REMOVE(NODE)

#define _LLD_FOREWARD_ACTION(H, T, INST)				      \
_MBEGIN									      \
    __XC_REGISTER LLD_TYPE(H) __node, ___next_node;			      \
									      \
    __LLD_SANITY_CHECK(H, T);						      \
									      \
    for (__node = __LLD_FIRST(H, T); __node != (T); __node = ___next_node)    \
    {									      \
	___next_node = __LLD_NEXT(__node);				      \
									      \
	INST								      \
    }									      \
_MEND

#define _LLD_BACKWARD_ACTION(H, T, INST)				      \
_MBEGIN									      \
    __XC_REGISTER LLD_TYPE(H) __node, ___prev_node;			      \
									      \
    __LLD_SANITY_CHECK(H, T);						      \
									      \
    for (__node = __LLD_LAST(H, T); __node != (H); __node = ___prev_node)     \
    {									      \
	___prev_node = __LLD_PREV(__node);				      \
									      \
	INST								      \
    }									      \
_MEND

#define _LLD_ACTION(H, T, INST)		_LLD_FOREWARD_ACTION(H, T, INST)
#define _LLD_FWD_ACTION(H, T, INST)	_LLD_FOREWARD_ACTION(H, T, INST)
#define _LLD_BWD_ACTION(H, T, INST)	_LLD_BACKWARD_ACTION(H, T, INST)

#define _LLD_CLEAN(H, T, INST)						      \
_MBEGIN									      \
    _LLD_ACTION(H, T, INST);						      \
									      \
    _LLD_INIT(H, T);							      \
_MEND

#define _LLD_PURGE(H, T, INST)	_LLD_PURGE(H, T, INST)

#define __LLD_LENGTH(H, T, LEN)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __i = 0;					      \
									      \
    _LLD_FOREACH_NODE(H, T, __i++;);					      \
									      \
    (LEN) = __i;							      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _LLD_LENGTH(H, T)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __len;						      \
									      \
    __LLD_LENGTH(H, T, __len);						      \
									      \
    __len;								      \
_MEND
#else
#define _LLD_LENGTH(H, T, LEN)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __len;						      \
									      \
    __LLD_LENGTH(H, T, __len);						      \
									      \
    (LEN) = __len;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _LLD_PLENGTH(H, T, LEN)	((LEN) = _LLD_LENGTH(H, T))
#else
#define _LLD_PLENGTH(H, T, LEN)	_LLD_LENGTH(H, T, LEN)
#endif

#define __LLD_SELECT(H, T, INDEX, NODE)					      \
_MBEGIN									      \
    __XC_REGISTER size_t const __i = (INDEX);				      \
    __XC_REGISTER size_t       __j = 0;					      \
									      \
    _LLD_FOREACH(H, T, NODE) if (__j++ == __i) break;			      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _LLD_SELECT(H, T, INDEX)					      \
_MBEGIN									      \
    __XC_REGISTER LLD_TYPE(H) __node;					      \
									      \
    __LLD_SELECT(H, T, INDEX, __node);					      \
									      \
    __node;								      \
_MEND
#else
#define _LLD_SELECT(H, T, INDEX, NODE)					      \
_MBEGIN									      \
    __XC_REGISTER LLD_TYPE(H) __node;					      \
									      \
    __LLD_SELECT(H, T, INDEX, __node);					      \
									      \
    (NODE) = __node;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _LLD_PSELECT(H, T, INDEX, NODE)	((NODE) = _LLD_SELECT(H, T, INDEX))
#else
#define _LLD_PSELECT(H, T, INDEX, NODE)	_LLD_SELECT(H, T, INDEX, NODE)
#endif


/*****************************************************************************
 * HIGH LEVEL MACROS
 *****************************************************************************/


#define _LLD_HEAD(LIST)	__LLD_HEAD((assert((LIST) != NULL), LIST))
#define _LLD_TAIL(LIST)	__LLD_TAIL((assert((LIST) != NULL), LIST))


#define LLD_INIT(LIST)				_LLD_INIT(_LLD_HEAD(LIST), _LLD_TAIL(LIST))
#define	LLD_NEXT(LIST, MARK)			_LLD_NEXT(_LLD_HEAD(LIST), _LLD_TAIL(LIST), MARK)
#define	LLD_PREV(LIST, MARK)			_LLD_PREV(_LLD_HEAD(LIST), _LLD_TAIL(LIST), MARK)
#define LLD_FIRST(LIST)				_LLD_FIRST(_LLD_HEAD(LIST), _LLD_TAIL(LIST))
#define LLD_LAST(LIST)				_LLD_LAST(_LLD_HEAD(LIST), _LLD_TAIL(LIST))
#define LLD_APPEND(LIST, NODE)			_LLD_APPEND(_LLD_HEAD(LIST), _LLD_TAIL(LIST), NODE)
#define LLD_INSERT(LIST, NODE)			_LLD_INSERT(_LLD_HEAD(LIST), _LLD_TAIL(LIST), NODE)
#define LLD_PREPEND(LIST, NODE)			_LLD_PREPEND(_LLD_HEAD(LIST), _LLD_TAIL(LIST), NODE)
#define LLD_INSERT_BEFORE(NODE, MARK)		_LLD_INSERT_BEFORE(NODE, MARK)
#define LLD_INSERT_AFTER(NODE, MARK)		_LLD_INSERT_AFTER(NODE, MARK)
#define LLD_REMOVE(NODE)			_LLD_REMOVE(NODE)
#define LLD_EXTRACT(NODE)			_LLD_EXTRACT(NODE)
#define LLD_FOREACH_NODE(LIST, INST)		_LLD_FOREACH_NODE(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_FOREACH(LIST, NODE)			_LLD_FOREACH(_LLD_HEAD(LIST), _LLD_TAIL(LIST), NODE)
#define LLD_FOREACH_COND(LIST, NODE, COND)	_LLD_FOREACH_COND(_LLD_HEAD(LIST), _LLD_TAIL(LIST), NODE, COND)
#define LLD_ACTION(LIST, INST)			_LLD_ACTION(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_FWD_ACTION(LIST, INST)		_LLD_FWD_ACTION(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_BWD_ACTION(LIST, INST)		_LLD_BWD_ACTION(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_FOREWARD_ACTION(LIST, INST)		_LLD_FOREWARD_ACTION(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_BACKWARD_ACTION(LIST, INST)		_LLD_BACKWARD_ACTION(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_CLEAN(LIST, INST)			_LLD_CLEAN(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)
#define LLD_PURGE(LIST, INST)			_LLD_PURGE(_LLD_HEAD(LIST), _LLD_TAIL(LIST), INST)

/*
 * Non portable versions
 */

#if defined __GNUC__ && !defined __STRICT_ANSI__

/* ====> Function behaviour */

#define LLD_LENGTH(LIST)			_LLD_LENGTH(_LLD_HEAD(LIST), _LLD_TAIL(LIST))
#define LLD_SELECT(LIST, I)			_LLD_SELECT(_LLD_HEAD(LIST), _LLD_TAIL(LIST), I)

#else

/* ====> strict ANSI conformance */

#define LLD_LENGTH(LIST, LEN)			_LLD_LENGTH(_LLD_HEAD(LIST), _LLD_TAIL(LIST), LEN)
#define LLD_SELECT(LIST, I, NODE)		_LLD_SELECT(_LLD_HEAD(LIST), _LLD_TAIL(LIST), I, NODE)

#endif

/*
 * Portable versions
 */

#define LLD_PLENGTH(LIST, LEN)			_LLD_PLENGTH(_LLD_HEAD(LIST), _LLD_TAIL(LIST), LEN)
#define LLD_PSELECT(LIST, I, NODE)		_LLD_PSELECT(_LLD_HEAD(LIST), _LLD_TAIL(LIST), I, NODE)


#endif /* XC LListD HEADER */


/*------------------------------/ END OF FILE /------------------------------*/
