/*  Copyright (C) 1996-2001  Martin Vicente			      -*- c -*-
    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*  XCLIB 0.3	LIST:		SINGLY LINKED LIST			     */


#ifndef __XC_List_H
#define __XC_List_H


/* Standard C Library: */
#include <stddef.h>
#include <assert.h>

/* eXtended C Library: */
#include <xc/version.h>		/* VERSION OF EXTENDED C LIBRARY	     */


/*****************************************************************************
 * CUSTOMIZATION PART
 *****************************************************************************/


#ifndef L_NodeRec
#define L_NodeRec   L_NodeRec
typedef struct node L_NodeRec;
#endif

#ifndef L_ListRec
#define L_ListRec   L_ListRec
typedef struct list L_ListRec;
#endif

#ifndef L_Node
#define L_Node     L_Node
typedef L_NodeRec *L_Node;
#endif

#ifndef L_List
#define L_List     L_List
typedef L_ListRec *L_List;
#endif


#ifndef L_next
#define L_next	next
#endif

#ifndef L_link
#define L_link	link
#endif


/*****************************************************************************
 * INTERNALS
 *****************************************************************************/


#if defined __GNUC__ && !defined __STRICT_ANSI__
#define L_TYPEREC(NODEREC)	typeof (NODEREC)
#define L_TYPE(NODE)		typeof (NODE)
#else
#define L_TYPEREC(NODEREC)	L_NodeRec
#define L_TYPE(NODE)		L_Node
#endif


#define	__L_NEXT(NODE)	((NODE)->L_next)
#define	__L_LINK(LIST)	((LIST)->L_link)


#define __L_FIRST(L)	(L)


#define __L_SANITY_CHECK(L) assert					      \
									      \
 ( 1 )


#define _L_INIT(L)							      \
_MBEGIN									      \
    (L) = NULL;								      \
_MEND

#define _L_FOREACH_COND(L, NODE, COND)					      \
									      \
 for (									      \
									      \
 __L_SANITY_CHECK(L),							      \
									      \
 (NODE) = __L_FIRST(L); (NODE) != NULL && (COND); (NODE) = __L_NEXT(NODE))

#define _L_FOREACH(L, NODE)	_L_FOREACH_COND(L, NODE, 1)

#define _L_FOREACH_NODE(L, INST)					      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) __node;					      \
									      \
    _L_FOREACH(L, __node) { INST }					      \
_MEND

#define _L_NEXT(NODE)							      \
									      \
 ( __L_NEXT(NODE) )

#define	__L_PREV(L, NODE, PREV)						      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) ____n;					      \
									      \
    (PREV) = (NULL);							      \
									      \
    _L_FOREACH_COND(L, ____n, ____n != (NODE)) (PREV) = ____n;		      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_L_PREV(L, NODE)						      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(NODE) ___prev;					      \
									      \
    __L_PREV(L, NODE, ___prev);						      \
									      \
    ___prev;								      \
_MEND
#else
#define	_L_PREV(L, NODE, PREV)						      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) ___prev;					      \
									      \
    __L_PREV(L, NODE, ___prev);						      \
									      \
    (PREV) = ___prev;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_L_PPREV(L, NODE, PREV)	((PREV) = _L_PREV(L, NODE))
#else
#define	_L_PPREV(L, NODE, PREV)	_L_PREV(L, NODE, PREV)
#endif

#define _L_FIRST(L)							      \
									      \
 (__L_SANITY_CHECK(L), __L_FIRST(L))

#define	__L_LAST(L, LAST)						      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(LAST) ____prev_node = NULL, ____node;		      \
									      \
    _L_FOREACH(L, ____node) ____prev_node = ____node;			      \
									      \
    (LAST) = ____prev_node;						      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_L_LAST(L)							      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) ___last_node;				      \
									      \
    __L_SANITY_CHECK(L);						      \
									      \
    __L_LAST(L, ___last_node);						      \
									      \
    ___last_node;							      \
_MEND
#else
#define	_L_LAST(L, LAST)						      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) ___last_node;				      \
									      \
    __L_SANITY_CHECK(L);						      \
									      \
    __L_LAST(L, ___last_node);						      \
									      \
    (LAST) = ___last_node;						      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define	_L_PLAST(L, LAST)	((LAST) = _L_LAST(L))
#else
#define	_L_PLAST(L, LAST)	_L_LAST(L, LAST)
#endif

#define _L_APPEND(L, N)							      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) ___last_node;				      \
									      \
    __L_SANITY_CHECK(L);						      \
									      \
    assert((N) != NULL);						      \
									      \
    __L_LAST(L, ___last_node);						      \
									      \
    if (___last_node != NULL)						      \
    {									      \
	__L_NEXT(___last_node) = (N);					      \
    }									      \
    else								      \
    {									      \
	(L) = (N);							      \
    }									      \
									      \
    __L_NEXT(N) = NULL;							      \
_MEND

#define _L_INSERT(L, N)							      \
_MBEGIN									      \
    __L_SANITY_CHECK(L);						      \
									      \
    assert((N) != NULL);						      \
									      \
    __L_NEXT(N) = (L);							      \
									      \
    (L) = (N);								      \
_MEND

#define _L_PREPEND(L, NODE) _L_INSERT(L, NODE)

#define _L_INSERT_BEFORE(L, NODE, MARK)					      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) ___prev_mark;				      \
									      \
    __L_SANITY_CHECK(L);						      \
									      \
    assert((NODE) != NULL);						      \
    assert((MARK) != NULL);						      \
									      \
    __L_NEXT(NODE) = (MARK);						      \
									      \
    __L_PREV(L, MARK, ___prev_mark);					      \
									      \
    if (___prev_mark != NULL)						      \
    {									      \
	__L_NEXT(___prev_mark) = (NODE);				      \
    }									      \
    else								      \
    {									      \
	(L) = (NODE);							      \
    }									      \
_MEND

#define _L_INSERT_AFTER(NODE, MARK)					      \
_MBEGIN									      \
    __L_SANITY_CHECK(L);						      \
									      \
    assert((NODE) != NULL);						      \
    assert((MARK) != NULL);						      \
									      \
    __L_NEXT(NODE) = __L_NEXT(MARK);					      \
									      \
    __L_NEXT(MARK) = (NODE);						      \
_MEND

#define __L_REMOVE(L, NODE)						      \
_MBEGIN									      \
    __L_SANITY_CHECK(L);						      \
									      \
    assert((NODE) != NULL);						      \
									      \
    if ((NODE) == (L))							      \
    {									      \
	(L) = __L_NEXT(L);						      \
    }									      \
    else								      \
    {									      \
	__XC_REGISTER L_TYPE(NODE) ___prev;				      \
									      \
	__L_PREV(L, NODE, ___prev);					      \
									      \
	__L_NEXT(___prev) = __L_NEXT(NODE);				      \
    }									      \
									      \
    __L_NEXT(NODE) = NULL;						      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _L_REMOVE(L, NODE)	(__L_REMOVE(L, NODE), (NODE))
#else
#define _L_REMOVE(L, NODE)	__L_REMOVE(L, NODE)
#endif

#define _L_EXTRACT(L, NODE) _L_REMOVE(L, NODE)

#define _L_FOREWARD_ACTION(L, INST)					      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) __node, ___next_node;			      \
									      \
    __L_SANITY_CHECK(L);						      \
									      \
    for (__node = __L_FIRST(L); __node != NULL; __node = ___next_node)	      \
    {									      \
	___next_node = __L_NEXT(__node);				      \
									      \
	INST								      \
    }									      \
_MEND

#define _L_BACKWARD_ACTION(L, INST)					      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) __node, ___prev_node;			      \
									      \
    __L_SANITY_CHECK(L);						      \
									      \
    __L_LAST(L, __node);						      \
									      \
    for (; __node != NULL; __node = ___prev_node)			      \
    {									      \
	__L_PREV(L, __node, ___prev_node);				      \
									      \
	INST								      \
    }									      \
_MEND

#define _L_ACTION(L, INST)	_L_FOREWARD_ACTION(L, INST)
#define _L_FWD_ACTION(L, INST)	_L_FOREWARD_ACTION(L, INST)
#define _L_BWD_ACTION(L, INST)	_L_BACKWARD_ACTION(L, INST)

#define _L_CLEAN(L, INST)						      \
_MBEGIN									      \
    _L_ACTION(L, INST);							      \
									      \
    _L_INIT(L);								      \
_MEND

#define _L_PURGE(L, INST)	_L_PURGE(L, INST)

#define __L_LENGTH(L, LEN)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __i = 0;					      \
									      \
    _L_FOREACH_NODE(L, __i++;);						      \
									      \
    (LEN) = __i;							      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _L_LENGTH(L)							      \
_MBEGIN									      \
    __XC_REGISTER size_t __len;						      \
									      \
    __L_LENGTH(L, __len);						      \
									      \
    __len;								      \
_MEND
#else
#define _L_LENGTH(L, LEN)						      \
_MBEGIN									      \
    __XC_REGISTER size_t __len;						      \
									      \
    __L_LENGTH(L, __len);						      \
									      \
    (LEN) = __len;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _L_PLENGTH(L, LEN)	((LEN) = _L_LENGTH(L))
#else
#define _L_PLENGTH(L, LEN)	_L_LENGTH(L, LEN)
#endif

#define __L_SELECT(L, INDEX, NODE)					      \
_MBEGIN									      \
    __XC_REGISTER size_t const __i = (INDEX);				      \
    __XC_REGISTER size_t       __j = 0;					      \
									      \
    _L_FOREACH(L, NODE) if (__j++ == __i) break;			      \
_MEND

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _L_SELECT(L, INDEX)						      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) __node;					      \
									      \
    __L_SELECT(L, INDEX, __node);					      \
									      \
    __node;								      \
_MEND
#else
#define _L_SELECT(L, INDEX, NODE)					      \
_MBEGIN									      \
    __XC_REGISTER L_TYPE(L) __node;					      \
									      \
    __L_SELECT(L, INDEX, __node);					      \
									      \
    (NODE) = __node;							      \
_MEND
#endif

#if defined __GNUC__ && !defined __STRICT_ANSI__
#define _L_PSELECT(L, INDEX, NODE)	((NODE) = _L_SELECT(L, INDEX))
#else
#define _L_PSELECT(L, INDEX, NODE)	_L_SELECT(L, INDEX, NODE)
#endif


/*****************************************************************************
 * HIGH LEVEL MACROS
 *****************************************************************************/


#define _L_LINK(LIST)	__L_LINK((assert((LIST) != NULL), LIST))


#define L_INIT(LIST)				_L_INIT(_L_LINK(LIST))
#define	L_NEXT(MARK)				_L_NEXT(MARK)
#define L_FIRST(LIST)				_L_FIRST(_L_LINK(LIST))
#define L_APPEND(LIST, NODE)			_L_APPEND(_L_LINK(LIST), NODE)
#define L_INSERT(LIST, NODE)			_L_INSERT(_L_LINK(LIST), NODE)
#define L_PREPEND(LIST, NODE)			_L_PREPEND(_L_LINK(LIST), NODE)
#define L_INSERT_BEFORE(LIST, NODE, MARK)	_L_INSERT_BEFORE(_L_LINK(LIST), NODE, MARK)
#define L_INSERT_AFTER(NODE, MARK)		_L_INSERT_AFTER(NODE, MARK)
#define L_REMOVE(LIST, NODE)			_L_REMOVE(_L_LINK(LIST), NODE)
#define L_EXTRACT(LIST, NODE)			_L_EXTRACT(_L_LINK(LIST), NODE)
#define L_FOREACH_NODE(LIST, INST)		_L_FOREACH_NODE(_L_LINK(LIST), INST)
#define L_FOREACH(LIST, NODE)			_L_FOREACH(_L_LINK(LIST), NODE)
#define L_FOREACH_COND(LIST, NODE, COND)	_L_FOREACH_COND(_L_LINK(LIST), NODE, COND)
#define L_ACTION(LIST, INST)			_L_ACTION(_L_LINK(LIST), INST)
#define L_FWD_ACTION(LIST, INST)		_L_FWD_ACTION(_L_LINK(LIST), INST)
#define L_BWD_ACTION(LIST, INST)		_L_BWD_ACTION(_L_LINK(LIST), INST)
#define L_FOREWARD_ACTION(LIST, INST)		_L_FOREWARD_ACTION(_L_LINK(LIST), INST)
#define L_BACKWARD_ACTION(LIST, INST)		_L_BACKWARD_ACTION(_L_LINK(LIST), INST)
#define L_CLEAN(LIST, INST)			_L_CLEAN(_L_LINK(LIST), INST)
#define L_PURGE(LIST, INST)			_L_PURGE(_L_LINK(LIST), INST)

/*
 * Non portable versions
 */

#if defined __GNUC__ && !defined __STRICT_ANSI__

/* ====> Function behaviour */

#define L_PREV(LIST, MARK)			_L_PREV(_L_LINK(LIST), MARK)
#define L_LAST(LIST)				_L_LAST(_L_LINK(LIST))
#define L_LENGTH(LIST)				_L_LENGTH(_L_LINK(LIST))
#define L_SELECT(LIST, I)			_L_SELECT(_L_LINK(LIST), I)

#else

/* ====> strict ANSI conformance */

#define L_PREV(LIST, MARK, PREV)		_L_PREV(_L_LINK(LIST), MARK, PREV)
#define L_LAST(LIST, LAST)			_L_LAST(_L_LINK(LIST), LAST)
#define L_LENGTH(LIST, LEN)			_L_LENGTH(_L_LINK(LIST), LEN)
#define L_SELECT(LIST, I, NODE)			_L_SELECT(_L_LINK(LIST), I, NODE)

#endif

/*
 * Portable versions
 */

#define L_PPREV(LIST, MARK, PREV)		_L_PPREV(_L_LINK(LIST), MARK, PREV)
#define L_PLAST(LIST, LAST)			_L_PLAST(_L_LINK(LIST), LAST)
#define L_PLENGTH(LIST, LEN)			_L_PLENGTH(_L_LINK(LIST), LEN)
#define L_PSELECT(LIST, I, NODE)		_L_PSELECT(_L_LINK(LIST), I, NODE)


#endif /* XC List HEADER */


/*------------------------------/ END OF FILE /------------------------------*/
