/*  XCLIB - a set of C functions for general purposes		      -*- c -*-
    Copyright (C) 1996-2001  Martin Vicente

    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*\
  |  Martin Vicente, 1996-2001
  |
  | Project : XCLIB (eXtended C Library)
  | Purpose : DYNAMIC MULTI-DIMENSIONAL ARRAY
  | Version : 0.3
  | Language: ANSI C
  | File    : dynarr/dm.c
  | Author  : Martin VICENTE <martin@vicente.org>
\*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*/


#define _XCLIB_INTERNALS

#define _FACILITY FACILITY(dynarr)

#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>

#include <xc/error.h>

#include "xc/dynarr.h"

#include "global.h"

typedef int DIM;

void *dmalloc(size_t const size, int const ndims, DIM const dim1, ...)
{
    void *new;

    va_list vargs;

    assert(size > 0 && ndims > 0 && dim1 > 0);

    va_start(vargs, dim1);
    new = vdmalloc(ndims, size, dim1, vargs);
    va_end(vargs);

    return new;
}

void *vdmalloc(size_t const size, int const ndims, DIM const dim1, va_list vargs)
{
    DIM  *dims;
    void *arr;

    register int n;

    assert(size > 0 && ndims > 0 && dim1 > 0);

    dims = malloc(ndims * sizeof (DIM));

    if (dims == NULL)
    {
	return NULL;
    }

    dims[0] = dim1;

    for (n = 1; n < ndims; ++n)
    {
	dims[n] = va_arg(vargs, DIM);

	assert(dims[n] > 0);
    }

    arr = dcalloc(size, ndims, dims);

    free(dims);

    return arr;
}

void dmfree(void *const arr)
{
    assert(arr != NULL);

    dcfree(arr);
}


/* End Of File */
