/*  XCLIB - a set of C functions for general purposes		      -*- c -*-
    Copyright (C) 1996-2001  Martin Vicente

    This file is part of the eXtended C Library (XCLIB).

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (see the file COPYING.LIB).
    If not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.		     */


/*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*\
  |  Martin Vicente, 1996-2001
  |
  | Project : XCLIB (eXtended C Library)
  | Purpose : EXTRA MATHEMATIC FUNCTIONS
  | Version : 0.3
  | Language: ANSI C
  | File    : math/mul.h
  | Author  : Martin VICENTE <martin@vicente.org>
\*|+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|*/


/*****************************************************************************\

Rappel:

  EDOM    Math argument out of domain of function
  ERANGE  Math result not representable

Attention:

  L'algorithme utilis ici pour la version "best" n'est pas encore la meilleur
  que l'on puisse envisager. Il sera propos dans une version future.

\*****************************************************************************/


#define CARD	unsigned int
#define L_MAX	UINT_MAX
#define U_MIN	0

#if !defined INTEGER && !defined FLOATING_POINT ||\
     defined INTEGER &&  defined FLOATING_POINT
#error "xclib: *** INTEGER or FLOATING_POINT!"
#endif

static TYPE fast(TYPE table[], CARD count);
static TYPE best(TYPE table[], CARD count);

TYPE FNAME(TYPE table[], CARD count)
{
    return (_math_flags & _mathFast ? fast : best)(table, count);
}

static TYPE fast(TYPE table[], CARD count)
{
    TYPE result = 1;

    register CARD i;

    assert(table != NULL && count > 0);

    for (i = 0; i < count; ++i)
    {
	TYPE const x = table[i];

	if (!(_math_flags & _mathNoRangeChecking))
	{
	    if (Abs(result) > 1)
	    {
		if (Abs(x) > 1 && Abs(x) > TYPE_MAX / Abs(result))
		{
		    errno = ERANGE;

		    return TYPE_MAX; /* overflow */
		}
	    }
#ifdef FLOATING_POINT
	    else if (Abs(result) < 1)
	    {
		if (Abs(x) < 1 && Abs(x) < TYPE_MIN / Abs(result))
		{
		    errno = ERANGE;

		    return TYPE_MIN; /* underflow */
		}
	    }
#endif
	}

	result *= x;
    }

#ifdef INTEGER
    if (Abs(result) == TYPE_MAX || Abs(result) == TYPE_MAX_NEG) errno = 0;
#endif
#ifdef FLOATING_POINT
    if (Abs(result) == TYPE_MAX || Abs(result) == TYPE_MIN)     errno = 0;
#endif

    return result;
}

typedef int (*compar_fn_t)(const void *, const void *);

static int qcmp(TYPE *x1, TYPE *x2);

static TYPE best(TYPE table[], CARD count)
{
    TYPE result;

    assert(table != NULL && count > 0);

#ifdef INTEGER

   /* La valeur absolue du plus petit entier ngatif est indfinie, car le
    * codage d'entier sign sur n bits offre un intervalle de valeurs allant de
    * -2^(n-1)  2^(n-1) - 1.
    */

    {
	register CARD i;

	i = 0;

	do
	{
	    if (table[i] == TYPE_MAX_NEG)
	    {
		errno = EDOM;

		return TYPE_MAX_NEG;
	    }
	}
	while (i++ != count);
    }

   /* Effectuer ce test juste aprs le tri qui suit serait bien plus rapide,
    * car il suffirait de ne tester que le dernier lment. Mais comme le tri
    * utilise la fonction de calcul de la valeur absolue, ceci n'est pas
    * possible pour la raison indique plus haut.
    */

#endif

    /* Trie des valeurs dans l'ordre croissant de leur valeur absolue */
    qsort(table, count, sizeof table[0], (compar_fn_t)qcmp);

    assert(TYPE_MIN > 0 && TYPE_MAX > 0);

    if (table[0] == 0) return 0;

    result = 1;

    {
	register CARD l = 0;
	register CARD u = count - 1;

	while (l <= u)
	{
	    if (Abs(result) > 1)
	    {
		TYPE const x = table[l];

		if (!(_math_flags & _mathNoRangeChecking))
		{
		    if (Abs(x) > 1 && Abs(x) > TYPE_MAX / Abs(result))
		    {
			errno = ERANGE;

			return TYPE_MAX; /* overflow */
		    }
		}

		result *= x;

		if (l == L_MAX) break;

		l++;
	    }
#ifdef FLOATING_POINT
	    else if (Abs(result) < 1)
	    {
		TYPE const x = table[u];

		if (!(_math_flags & _mathNoRangeChecking))
		{
		    if (Abs(x) < 1 && Abs(x) < TYPE_MIN / Abs(result))
		    {
			errno = ERANGE;

			return TYPE_MIN; /* underflow */
		    }
		}

		result *= x;

		if (u == U_MIN) break;

		u--;
	    }
#endif
	    else /* Abs(result) == 1 */
	    {
		result *= table[u];

		if (u == U_MIN) break;

		u--;
	    }
	}
    }

#ifdef INTEGER
    if (Abs(result) == TYPE_MAX || Abs(result) == TYPE_MAX_NEG) errno = 0;
#endif
#ifdef FLOATING_POINT
    if (Abs(result) == TYPE_MAX || Abs(result) == TYPE_MIN)     errno = 0;
#endif

    return result;
}

static int qcmp(TYPE *x1, TYPE *x2)
{
    return Abs(*x1) > Abs(*x2) ? +1 : Abs(*x1) < Abs(*x2) ? -1 : 0;
}


/* END OF FILE */
