#define _XCLIB_SOURCE

#include <errno.h>
#include <float.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "xc/macros.h"
#include "xc/timer.h"
#include "xc/xalloc.h"
#include "xc/math.h"

#define TYPE	float

#define WITH_TIMER 1

#if WITH_TIMER
typedef const struct timer *Timer;
#endif

static void print_fresult(TYPE result);

int main()
{
#if 1
    TYPE *table;
    int   count;
#else
    TYPE  table[] = { -FLT_MAX, -FLT_EPSILON };
    TYPE  count   = cardinal(table);
#endif
    TYPE  value;
#if WITH_TIMER
    Timer timer;
#endif
    int   i;

#if 1
    count = 1000;
    table = xmalloc(count * sizeof (TYPE));

    for (i = 0; i < count; i++) table[i] = 1;
#endif

#if 1
    _math_flags  = 0;
    _math_flags |= _mathFast;
    _math_flags |= _mathNoRangeChecking;
#endif

#if WITH_TIMER
    timer = timer_ctl(NULL, timerNEW | timerSET);

    for (i = 0; i < 1000; ++i)
    {
#endif
	value = faddf(table, count);

#if WITH_TIMER
	timer_ctl(timer, timerTOP);
    }

    timer_ctl(timer, timerPUT | timerDEL);
#endif

    if ((value == +FLT_MAX || value == -FLT_MAX) && errno)
	puts(strerror(errno));
    else
	print_fresult(value);

    return 0;
}

static void print_fresult(TYPE result)
{
    if (result > 0)
    {
	if (result == HUGE_VAL)
	    printf("result = Infinity\n");
	else if (result == FLT_MAX)
	    printf("result = FLT_MAX\n");
	else
	    printf("result = %.20g\n", (double)result);
    }
    else if (result < 0)
    {
	if (result == -HUGE_VAL)
	    printf("result = -Infinity\n");
	else if (result == -FLT_MAX)
	    printf("result = -FLT_MAX\n");
	else
	    printf("result = %.20g\n", (double)result);
    }
    else
    {
	printf("result = ZERO\n");
    }
}
